# Copyright © 2024, UChicago Argonne, LLC
# BSD OPEN SOURCE LICENSE. Full license can be found in LICENSE
# Based on AequilibraE's code for bandwidth map creation
import sys
from math import ceil
from os.path import dirname, join
from random import randint

import qgis
from polaris.runs.run_utils import get_output_dirs
from PyQt5.QtCore import Qt
from qgis.core import QgsDataSourceUri, QgsProject, QgsStyle, QgsVectorLayer
from qgis.PyQt import uic
from qgis.PyQt.QtGui import QColor
from qgis.PyQt.QtWidgets import QDialog

from .map_slider import MapSlider
from .traffic_results import TrafficResults, base_interval_results as base_interval
from ..common_tools import running_on_ci

sys.modules["qgsfieldcombobox"] = qgis.gui
sys.modules["qgscolorbutton"] = qgis.gui
sys.modules["qgsmaplayercombobox"] = qgis.gui
FORM_CLASS, _ = uic.loadUiType(join(dirname(__file__), "forms/ui_bandwidths.ui"))

start_thickness_field = "ab_link_out_volume"
start_color_field = "ab_link_travel_delay"


class LinkMapCreationDialog(QDialog, FORM_CLASS):
    def __init__(self, qgis_project):
        QDialog.__init__(self)
        self.slider: MapSlider = None

        self._PQgis = qgis_project
        self._p = qgis_project.network

        self.iface = qgis_project.iface
        self.setupUi(self)

        self.tot_bands = 0
        self.htodo = 6
        self.scale = {"width": 10, "spacing": 0, "max_flow": 400}

        self.band_size = 10.0
        self.space_size = 0.01
        # self.drive_side = "right"
        self.link_layer = self._PQgis.layers.get("link", [None, None])[0]
        QgsProject.instance().addMapLayer(self.link_layer)
        self.results = TrafficResults(qgis_project, htodo=self.htodo)
        self.color_results = TrafficResults(qgis_project, htodo=self.htodo)
        self.load_link_layer()

        config = qgis_project.polaris_project.run_config
        config.data_dir = qgis_project.polaris_project.model_path
        iterations = ["root"] + [str(x.name) for x in get_output_dirs(config)]
        self.cob_base.addItems(iterations)
        self.cob_alternative.addItems(iterations)

        self.interval_bar.valueChanged.connect(self.change_range)

        self.chb_ramp.toggled.connect(self.deal_with_colors)
        self.chb_compare.toggled.connect(self.set_compare)
        self.chb_thickness.toggled.connect(self.set_thickness)
        self.grouping_slider.valueChanged.connect(self.change_grouping)

        self.but_run.clicked.connect(self.add_bands_to_map)

        self.random_rgb()
        myStyle = QgsStyle().defaultStyle()
        color_ramps = myStyle.colorRampNames()
        for i in color_ramps:
            self.cob_colorRamp.addItem(i)
        self.deal_with_colors()
        self.set_compare()
        self.setFixedHeight(418)
        self.change_grouping()
        self.add_fields_to_cboxes()

    def change_grouping(self):
        self.grouping_slider.setValue(int(base_interval * ceil(self.grouping_slider.value() / base_interval)))
        self.to_group.setText(f"{self.grouping_slider.value()} min")
        self.htodo = min(24, 3 * self.grouping_slider.value())
        self.results.htodo = self.htodo
        self.color_results.htodo = self.htodo
        self.interval_bar.setEnabled(self.htodo < 24)
        self.interval_bar.setPageStep(self.htodo * 2 if self.htodo < 24 else 1000)
        self.change_range()
        self.add_fields_to_cboxes()

    def load_link_layer(self):
        if self.link_layer is None:
            uri = QgsDataSourceUri()
            uri.setDatabase(self._PQgis.supply_path)
            uri.setDataSource("", "link", "geo")
            self.link_layer = QgsVectorLayer(uri.uri(), "link", "spatialite")
            self._PQgis.layers["link"] = [self.link_layer, self.link_layer.id()]
            QgsProject.instance().addMapLayer(self.link_layer)

    def deal_with_colors(self):
        self.box_color_ramp.setVisible(self.chb_ramp.isChecked())
        self.box_color_ramp.setEnabled(self.chb_ramp.isChecked())

        show_color = not self.chb_compare.isChecked() and not self.chb_ramp.isChecked()
        self.lab_color.setVisible(show_color)
        self.lab_color.setEnabled(show_color)

        self.mColorButton.setVisible(show_color)
        self.mColorButton.setEnabled(show_color)

        if not self.chb_ramp.isChecked() and not self.chb_compare.isChecked():
            self.chb_compare.setVisible(True)

    def set_compare(self):
        self.comparebox.setVisible(self.chb_compare.isChecked())
        self.box_color_ramp.setVisible(not self.chb_compare.isChecked())
        self.chb_ramp.setEnabled(not self.chb_compare.isChecked())
        if self.chb_compare.isChecked():
            self.chb_ramp.setChecked(False)
        else:
            self.deal_with_colors()

    def set_thickness(self):
        self.box_thickness.setVisible(self.chb_thickness.isChecked())
        if not self.chb_thickness.isChecked():
            self.chb_compare.setChecked(False)
            self.chb_ramp.setChecked(True)

    def add_fields_to_cboxes(self):
        tables_available = self.results.tables_for_aggregation(self.grouping_slider.value())
        color_metric = None if self.FieldColor.currentText() == "" else self.FieldColor.currentText()
        self.FieldColor.clear()
        self.FieldColor.addItems(tables_available)

        if color_metric is not None and color_metric in tables_available:
            self.FieldColor.setCurrentText(color_metric)

        thickness_metric = None if self.FieldThickness.currentText() == "" else self.FieldThickness.currentText()
        self.FieldThickness.clear()
        self.FieldThickness.addItems(tables_available)

        if thickness_metric is not None and thickness_metric in tables_available:
            self.FieldThickness.setCurrentText(thickness_metric)

    def add_bands_to_map(self):
        do_thickness = self.chb_thickness.isChecked()
        do_color = self.chb_ramp.isChecked()

        if do_thickness + do_color == 0:
            self.iface.messageBar().pushMessage("Error", "Doing nothing? Try again", level=3, duration=5)
            return

        if self.chb_compare.isChecked():
            base = self.cob_base.currentText()
            alt = self.cob_alternative.currentText()

            if base == alt:
                self.iface.messageBar().pushMessage(
                    "Error", "Comparing something to itself is silly. Try again", level=3, duration=5
                )
                return

            self.results = TrafficResults(qgis_project=self._PQgis, base=base, alternative=alt, htodo=self.htodo)

        self.but_run.setEnabled(False)

        if do_thickness:
            self.results.build_metric_layer(
                self.FieldThickness.currentText(), self.interval_bar.value(), self.grouping_slider.value()
            )

        if do_color:
            self.color_results.build_metric_layer(
                self.FieldColor.currentText(), self.interval_bar.value(), self.grouping_slider.value()
            )
        self.exit_procedure()

    def change_range(self):
        end_interval = min(24, self.interval_bar.value() + self.results.htodo)
        start_interval = end_interval - self.htodo

        if start_interval != self.interval_bar.value():
            self.interval_bar.setValue(start_interval)

        self.from_time.setText(f"{start_interval}:00")
        self.to_time.setText(f"{end_interval}:00")

    def random_rgb(self):
        rgb = [randint(0, 255) for _ in range(3)]
        a = QColor()
        a.setRgb(rgb[0], rgb[1], rgb[2])
        self.mColorButton.setColor(a)

    def exit_procedure(self):
        from_interval = int(self.interval_bar.value() * 3600 / self.results.step)
        self.slider = MapSlider(
            self._PQgis,
            self.results,
            self.color_results,
            from_interval,
            self.cob_colorRamp.currentText(),
            self.chb_invert.isChecked(),
            self.chb_thickness.isChecked(),
            self.chb_ramp.isChecked(),
            self.mColorButton.color(),
            self.chb_compare.isChecked(),
        )

        self.slider.setWindowFlags(Qt.WindowStaysOnTopHint)
        self.close()
        if running_on_ci():
            return
        self.slider.show()
        self.slider.exec_()
